cmp_str = {
    'ORTH' :    lambda x : x.str,
    'TEXT' :    lambda x : x.str,
    'LOWER' :    lambda x : x.str.lower(),
    'LEMMA' :    lambda x : x.lemma,
    'POS' :      lambda x : x.pos,
    'DEP' :      lambda x : x.dep,
    'ENT_TYPE' : lambda x : x.ner.get('type', [''])[0],
    'CATEGORY_SUPER' : lambda x : [k[0] for k in x.meaning],
    'CATEGORY_SUB' : lambda x : [k[1] for k in x.meaning],
}

cmp_int = {
    'LENGTH' :   lambda x : len(x.str)
}

cmp_bool = {
    'IS_ALPHA':     lambda x: x.str.isalpha(),
    'IS_ASCII':     lambda x: x.str.isascii(),
    'IS_DIGIT':     lambda x: x.str.isdigit(),
    'IS_LOWER':     lambda x: x.str.islower(),
    'IS_UPPER':     lambda x: x.str.isupper(),
    'IS_TITLE':     lambda x: x.str.istitle(),
    'IS_PUNCT':     lambda x: True if x in ['.', '!', '?', ';'] else False,
    'IS_SPACE':     lambda x: x.str.isspace(),
    'IS_SENT_START': lambda x: x.idx == 0,
    'LIKE_NUM':     lambda x: x.pos == 'CD',
    'LIKE_URL':     lambda x: 'url' in x.ner.get('type', []),
    'LIKE_EMAIL':   lambda x: 'mail' in x.ner.get('type', []),
}

quantifiers = {
    "!":lambda x : True,
    "?":lambda x : True,
    "+":lambda x : True,
    "*":lambda x : True,
}

ops = {
    'IN':     lambda x, lst: x in lst if not isinstance(x, list) else ops['ISSUPERSET'](x, lst),
    'NOT IN': lambda x, lst: x not in lst if not isinstance(x, list) else not ops['ISSUPERSET'](x, lst),
    'ISSUBSET': lambda x_lst, lst: list(set(x_lst) & set(lst)) == x_lst and (x_lst or not lst),
    'ISSUPERSET': lambda x_lst, lst: list(set(x_lst) & set(lst)) == lst,
    '==': lambda x, y: x == y,
    '!=': lambda x, y: x != y,
    '>=': lambda x, y: x >= y,
    '<=': lambda x, y: x <= y,
    '>':  lambda x, y: x > y,
    '<':  lambda x, y: x < y,
}

types_check = {
    "IN": [list],
    "NOT IN": [list],
    "ISSUBSET": [list],
    "ISSUPERSET": [list],
}

valid_keys = list(cmp_str.keys()) + list(cmp_bool.keys()) + list(cmp_int.keys()) + list(quantifiers.keys()) + ['OP']

def compare_attr(token, node):
    attributes = [k for k in node.keys() if k != 'OP']
    for attribute in attributes:
        if attribute not in valid_keys:
            raise Exception("Error attribute " + attribute +  " does not exist.")
    for attribute in attributes:
        if isinstance(node[attribute], list):
            node_ops = ['IN']
            node[attribute] = {'IN':node[attribute]}
        elif not isinstance(node[attribute], dict):
            node_ops = ['==']
            node[attribute] = {'==':node[attribute]}
        else:
            node_ops = node[attribute].keys()
        if cmp_str.get(attribute, None):
            res = cmp_str[attribute](token)
            for op in node_ops:
                if op == '==' and not isinstance(node[attribute][op], str):
                    raise Exception(attribute + " attribute expects a string value.")
                elif types_check.get(op, '') and type(node[attribute][op]) not in types_check[op]:
                    raise Exception(attribute + " attribute and operator " + op + " expect one of " + str(types_check[op]))
                if not ops[op](res, node[attribute][op]):
                    return False
        elif cmp_int.get(attribute, None):
            res = cmp_int[attribute](token)
            for op in node_ops:
                if op == '==' and not isinstance(node[attribute][op], int):
                    raise Exception(attribute + " attribute expects an integer value.")
                elif types_check.get(op, '') and type(node[attribute][op]) not in types_check[op]:
                    raise Exception(attribute + " attribute and operator " + op + " expect one of " + str(types_check[op]))
                if not ops[op](res, node[attribute][op]):
                    return False
        elif cmp_bool.get(attribute, None):
            if not isinstance(node[attribute].get('==', None), bool):
                raise Exception(attribute + " attribute expects a boolean value.")
            if not cmp_bool[attribute](token) == node[attribute]['==']:
                return False
    return True


def addLem(lem, lemToAdd):
    if lemToAdd :
        if lem == '_':
            return lemToAdd
        else :
            return lem + '|' + lemToAdd


def transform_data(data):
    res = []
    if not isinstance(data, list):
        data = [data]
    for d in data:
        lem = '_'
        w = []
        w.append(d['source'])
        if 'lemmatizer' not in d.keys():
            w.append(d['source'])
        elif type(d["lemmatizer"]) is dict:
            if "infinit" in d["lemmatizer"].keys():
                w.append(d['lemmatizer']['infinit'])

            elif "lemma" in d["lemmatizer"].keys():
                if d['lemmatizer']["lemma"] != "":
                    w.append(d['lemmatizer']['lemma'])
                else:
                    w.append(d['source'])

            else:
                w.append(d['source'])

            lemmatizer = d['lemmatizer']


        elif type(d['lemmatizer']) is list:
            if 'infinit' in d['lemmatizer'][0].keys():
                w.append(d['lemmatizer'][0]['infinit'])

            elif 'infinit' in d['lemmatizer'][1].keys():
                w.append(d['lemmatizer'][1]['infinit'])

            else:
                w.append(d['source'])

            lemmatizer = d['lemmatizer'][0]


        else:
            w.append(d['source'])

        if 'gender' in lemmatizer.keys() :
            try:
                if lemmatizer['gender']['female']==True:
                    gender = 'Gender=Fem'
                elif lemmatizer['gender']['female']==False:
                    gender = 'Gender=Masc'
                lem = addLem(lem, gender)
            except:
                pass
                # print("Update 20191105 - CCO add exception : No female is present in key gender for word '{}'. Key = {}".format(d["source"], lemmatizer))

            try:
                if lemmatizer['gender']['plural']==True:
                    number = 'Number=Plur'
                elif lemmatizer['gender']['plural']==False:
                    number = 'Number=Sing'
                lem = addLem(lem, number)
            except:
                pass
                # print("Update 20191105 - CCO add exception : No plural is present in key gender for word '{}'. Key = {}".format(d["source"], lemmatizer))

            #Disjunct cases, not yes exhaustive. Should it be ?? Check in code how it is treated.
        if 'conjugate' in lemmatizer.keys() :
            if len(lemmatizer['conjugate']) == 1:
                try:
                    mode = lemmatizer['conjugate'][0]['mode']
                    if mode == 'indicative':
                        mood = 'Mood=Ind'
                    elif mode == 'subjonctive':
                        mood = 'Mood=Sub'
                    elif mode == 'conditional':
                        mood = 'Mood=Cnd'
                    elif mode == 'imperative':
                        mood = 'Mood=Imp'
                    elif mode == 'infinitive':
                        mood = 'VerbForm=Inf'
                    lem = addLem(lem, mood)
                except:
                    pass
                    # print("Update 20191105 - CCO add exception : No mood is present in key conjugate for word '{}'. Key = {}".format(d["source"], lemmatizer['conjugate']))
                try:
                    person = "Person=" + str(lemmatizer['conjugate'][0]['pronom'])
                    lem = addLem(lem, person)
                except:
                    pass
                    # print("Update 20191105 - CCO add exception : No person is present in key conjugate for word '{}'. Key = {}".format(d["source"], lemmatizer['conjugate']))

                try:
                    if int(lemmatizer['conjugate'][0]['pronom']) in [1, 2, 3]:
                        number = "Number=Sing"
                    elif int(lemmatizer['conjugate'][0]['pronom']) in [4, 5, 6]:
                        number = "Number=Plur"
                    lem = addLem(lem, number)
                except:
                    pass
                    # print("Update 20191105 - CCO add exception : No number is present in key conjugate for word '{}'. Key = {}".format(d["source"], lemmatizer['conjugate']))


                # Analogically with conjugate, we might be supposed to disjunct cases

                try:
                    temps = str(lemmatizer['conjugate'][0]['temps'])
                    if temps == 'present':
                        tense = "Tense=Pres"
                    elif temps == '':
                        tense = "Tense=Imp"
                    elif temps == 'past':
                        tense = "Tense=Past"
                    elif temps == 'future':
                        tense = "Tense=Fut"
                    lem = addLem(lem, tense)
                except:
                    pass
                    # print("Update 20191105 - CCO add exception : No tense is present in key conjugate for word '{}'. Key = {}".format(d["source"], lemmatizer['conjugate']))

                try:
                    if mood:
                        verbform = "VerbForm=Fin"
                        lem = addLem(lem, verbform)
                except:
                    pass
                    # print("Update 20191105 - CCO add exception : No mood is present in key conjugate for word '{}'. Key = {}".format(d["source"], lemmatizer['conjugate']))



        if 'mode' in lemmatizer.keys() :
            mode = lemmatizer['mode']
            try:
                if mode == 'define':
                    definite = 'Definite=Def'
                    lem = addLem(lem, definite)
                elif mode == 'undefine':
                    definite = 'Definite=Ind'
                    lem = addLem(lem, definite)
            except:
                pass
                # print("Update 20191105 - CCO add exception : No definite is present in key mode for word '{}'. Key = {}".format(d["source"], lemmatizer))


        elif 'lemma' in lemmatizer.keys():
            if lemmatizer['lemma'] in ['un', 'une']:
                definite = 'Definite=Ind'
                try:
                    lem = addLem(lem, definite)
                except:
                    pass
                    # print("Update 20191105 - CCO add exception : No definite is present in key definite for word '{}'. Key = {}".format(d["source"], lemmatizer))


        if 'tag' in d.keys():
            tag = d['tag']
            if tag=='D' or tag=='PD' :
                if 'mode' in lemmatizer.keys():
                    if lemmatizer['mode'] == 'demonstrative':
                        prontype = 'PronType=Dem'
                    elif lemmatizer['mode'] == 'possessive':
                        prontype = 'Poss=Yes|PronType=Pos'
                    elif lemmatizer['mode'] == 'define':
                        prontype = 'PronType=Art'
                    elif 'pronom' in lemmatizer.keys():
                        try:
                            person = "Person=" + str(lemmatizer['pronom'])
                            lem = addLem(lem, person)
                        except:
                            pass
                            # print("Update 20191105 - CCO add exception : No person is present in key conjugate for word '{}'. Key = {}".format(d["source"], lemmatizer))

                        try:
                            number = "Number=Sing" if int(lemmatizer['pronom']) <= 3 else "Number=Plur"
                            lem = addLem(lem, number)
                        except:
                            pass
                            # print("Update 20191105 - CCO add exception : No number is present in key conjugate for word '{}'. Key = {}".format(d["source"], lemmatizer))

                        prontype = 'PronType=Prs'
                    try:
                        lem = addLem(lem, prontype)
                    except:
                        pass
                        # print("Update 20191105 - CCO add exception : No prontype is present in key conjugate for word '{}'. Key = {}".format(d["source"], lemmatizer))
                elif 'category' in lemmatizer.keys():
                    if lemmatizer['category'] == 'negation':
                        prontype = 'PronType=Neg'
            if tag == "VP" :
                try:
                    lem = addLem(lem, "VerbForm=Part")
                    lem = addLem(lem, "Tense=Past")
                except:
                    pass
                    # print("Update 20191112 - CCO add exception")

            if tag == "CLO":
                print(lemmatizer)
                if lemmatizer["lemma"] == "se":
                    lem = addLem(lem, "Reflex=Yes")
                    if d["source"] == "se":
                        lem = addLem(lem, "Person=3")
                    if d["source"] == "me":
                        lem = addLem(lem, "Person=1")
                        lem = addLem(lem, "Number=Sing")
                    if d["source"] == "te":
                        lem = addLem(lem, "Person=2")
                        lem = addLem(lem, "Number=Sing")
                    if d["source"] == "nous":
                        lem = addLem(lem, "Person=1")
                        lem = addLem(lem, "Number=Plur")
                    if d["source"] == "vous":
                        lem = addLem(lem, "Person=2")
                        lem = addLem(lem, "Number=Plur")

                if lemmatizer["lemma"] == "le":
                    if d["source"] == "me":
                        lem = addLem(lem, "Person=1")
                        lem = addLem(lem, "Number=Sing")
                    if d["source"] == "le":
                        lem = addLem(lem, "Person=3")
                        lem = addLem(lem, "Number=Sing")
                    if d["source"] == "les":
                        lem = addLem(lem, "Person=3")
                        lem = addLem(lem, "Number=Plur")

                if d["source"] == "y":
                    lem = addLem(lem, "Person=3")

            if tag == "CLS":
                if d["source"] == "je":
                    lem = addLem(lem, "Number=Sing")
                    lem = addLem(lem, "Person=1")
                elif d["source"] == "tu":
                    lem = addLem(lem, "Number=Sing")
                    lem = addLem(lem, "Person=2")
                elif d["source"] == "il":
                    lem = addLem(lem, "Number=Sing")
                    lem = addLem(lem, "Person=3")
                    lem = addLem(lem, "Gender=Masc")
                elif d["source"] == "elle":
                    lem = addLem(lem, "Number=Sing")
                    lem = addLem(lem, "Person=3")
                    lem = addLem(lem, "Gender=Fem")
                elif d["source"] == "on":
                    lem = addLem(lem, "Number=Sing")
                    lem = addLem(lem, "Person=3")
                elif d["source"] == "nous":
                    lem = addLem(lem, "Number=Plur")
                    lem = addLem(lem, "Person=1")
                elif d["source"] == "vous":
                    lem = addLem(lem, "Number=Plur")
                    lem = addLem(lem, "Person=2")
                elif d["source"] == "ils":
                    lem = addLem(lem, "Number=Plur")
                    lem = addLem(lem, "Person=3")
                    lem = addLem(lem, "Gender=Masc")
                elif d["source"] == "elles":
                    lem = addLem(lem, "Number=Plur")
                    lem = addLem(lem, "Person=3")
                    lem = addLem(lem, "Gender=Fem")

        if 'category' in lemmatizer.keys():
            if lemmatizer['category'] == 'negation':
                polarity = 'Polarity=Neg'
                try:
                    lem = addLem(lem, polarity)
                except:
                    pass
                    # print("Update 20191105 - CCO add exception : No polarity is present in key conjugate for word '{}'. Key = {}".format(d["source"], lemmatizer))
        res.append(lem)
    return(res)



from .attribute_checker import compare_attr

class DepTree:
    def __init__(self, sentence_data, idx, root, parent_ref = None):
        self.data = sentence_data
        self.root = self if root is None else root
        self.idx = idx
        self.parent_ref = parent_ref
        self.children = []
        self.grow_children()
        self.tokens = None
    
    def grow_children(self):
        for i, item in enumerate(self.data):
            if item['ref'] == self.idx:
                self.children.append(DepTree(self.data, i, self.root, self))
    
    def __repr__(self):
        return self.str

    def __str__(self):
        return self.str

    def print_tree(self, depth = 0, last = False):
        phrase = ""
        for _ in range(depth - 1):
            phrase += '|\t' if last == False else '\t'
        if depth > 0:
            phrase += '|______'
        phrase += "{} [{}]({}) {} {}".format(self.str, self.idx, self.root.tokens[self.idx].dep, \
                                    self.root.tokens[self.idx].pos, self.root.tokens[self.idx].dep)
        phrase += '\n'
        if self.children:
            for _, child in enumerate(self.children):
                phrase += child.print_tree(depth = depth + 1, last = False)
        if self.parent_ref == None:
            print(phrase)
        else:
            return (phrase)

    @property
    def str(self):
        return self.root.tokens[self.idx].str

    @property
    def child(self):
        return self.children
    
    @property
    def descendants(self):
        res = [self]
        for child in self.children:
            res += child.descendants
        return res
    
    @property
    def parent(self):
        return self.parent_ref

    @property
    def parents(self):
        parents = []
        tmp = self
        while tmp.parent:
            parents.append(tmp.parent)
            tmp = tmp.parent
        return parents

    @property
    def siblings(self):
        return self.parent.children if self.parent else []
    
    @property
    def left_siblings(self):
        return [child for child in self.parent.children if child.idx < self.idx]
    
    @property
    def right_siblings(self):
        return [child for child in self.parent.children if child.idx > self.idx]

    @property
    def prev(self):
        if self.idx >= 1:
            return self.root.find_node(self.idx - 1)
        else:
            return None
    
    @property
    def prevs(self):
        res = []
        for i in range(0, self.idx):
            res.append(self.root.find_node(i))
        return res

    @property
    def next(self):
        if self.idx < len(self.root.tokens):
            return self.root.find_node(self.idx + 1)
        else:
            return None

    @property
    def nexts(self):
        res = []
        for i in range(self.idx + 1, len(self.root.tokens)):
            res.append(self.root.find_node(i))
        return res

    def find_node(self, idx):
        if self.idx == idx:
            return self
        for child in self.children:
            tmp = child.find_node(idx)
            if tmp != None:
                return tmp 
        return None     

    def find_node_attribute(self, pattern):
        ret = []
        if compare_attr(self.root.tokens[self.idx], pattern):
            ret.append(self)
        for child in self.children:
            tmp = child.find_node_attribute(pattern)
            if tmp != None:
                ret += tmp
        return ret     


    @classmethod
    def grow_tree(self, sentence_data):
        root = None
        for i, item in enumerate(sentence_data['synthesis']):
            if item['dep'] == 'root':
                root = DepTree(sentence_data['synthesis'], i, None, None)
                break
        return root


import sys
import os
import json
import jsonlines as jsonl
from .client import Client
from .subtypes import Token, Subsentence
from .utils import flatten_lst, StrProperty, ListProperty, DictProperty
from .TextChunk import TextChunk

#Accepted arguments for hierarchical levels
GLOBAL =    ['g', 'global', 'glob']
DOC =       ['d', 'doc', 'document', 'documents']
SENT =      ['s', 'sentence', 'sent', 'sentences']
SUB =       ['sub', 'subsentence', 'subsentences']
TOK =       ['t', 'token', 'tok', 'tokens']

class PipelineError(Exception): pass
class RequestError(Exception): pass

def clear_data(data_json):
    def clean_recursif(node):
        if isinstance(node, list):
            for e in node:
                clean_recursif(e)  
        elif isinstance(node, dict):
            keys = list(node.keys())
            for k in keys:
                if k == 'confidence' or node[k] in [[], {}, None]:
                    node.pop(k)
                elif isinstance(node[k], dict):
                    clean_recursif(node[k])
                    if not node[k]:
                        node.pop(k)
                elif isinstance(node[k], list):
                    clean_recursif(node[k])
                    if not node[k]:
                        node.pop(k)
        else:
            return None

    data_json = {k:v for k,v in data_json.items() if v and k in ['source', 'ml_sentiment', 'proposition', 'sentiment', 'sentence_acts', 'ml_emotion', 'synthesis']}
    data_json['synthesis'] = [{k:v for k,v in i.items() if v not in [[], {}, None]} for i in data_json['synthesis']]
    clean_recursif(data_json)
    return data_json

class Sentence(TextChunk):
    __slots__ = ("data", "n", "max", "id")

    def __init__(self, data_sentence, idx=0):
        super(Sentence, self).__init__()
        self.data = clear_data(data_sentence)
        self.max = len(self.data.get('synthesis', []))
        self._ner_fix()
        self.id = idx

    # To modify when desambiguisation is active and only one NER entity is returned for each token
    def _ner_fix(self):
        for i, d in enumerate(self.data.get('synthesis', [])):
            ner = []
            for m in d.get('meaning', []):
                if 'super' in m and m['super']:
                    if m['super'].lower() in ['location', 'person', 'organization']:
                        if m['super'].lower() not in ner:
                            ner.append(m['super'].lower())
                    elif m['super'] == 'ENTITY':
                        ner.append(m['sub'])
                if 'sub' in m and m['sub']:
                    if m['sub'] == 'number':
                        ner.append(m['sub'])
            if ner:
                self.data['synthesis'][i]['type'] = ner

    def __repr__(self):
        return self.str

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < self.max:
            self.n += 1
            return Token(self.data.get('synthesis', [])[self.n - 1], self.n -1)
        else:
            raise StopIteration

    @property
    def idx(self):
        return self.id

    @ListProperty
    def sentences(self):
        return [self]

    def _get_subsentence(self, id, idx):
        ''' Dividing information for each subsentence, then it can use 
        normal Sentence methods to access its data'''
        data = {}
        data['synthesis'] = self.data['synthesis'][idx['start_id']:idx['end_id'] + 1]
        data['source'] = ' '.join([t.get('source', '') for t in data['synthesis']])
        if len(self.data['sentiment'].get('subsentences', [])) > id:
            data['sentiment'] = self.data['sentiment'].get('subsentences', [])[id]
        if len(self.data.get('ml_emotion', {}).get('subsentence', [])) > id:
            data['ml_emotion'] = {'sentence': self.data.get('ml_emotion', {}).get('subsentence', [])[id]}
        if len(self.data.get('ml_sentiment', {}).get('subsentence', [])) > id:
            data['ml_sentiment'] = {'sentence': self.data['ml_sentiment'].get('subsentence', [])[id]}
        return Subsentence(data)

    @ListProperty
    def tokens(self):
        return [Token(s, i) for i, s in enumerate(self.data.get('synthesis', []))]

    @ListProperty
    def subsentences(self):
        return [self._get_subsentence(id, idx) for id, idx in enumerate(self.data.get('proposition', []))]

    @ListProperty
    def token(self):
        return [s.get('source', None) for s in self.data.get('synthesis', [])]

    @StrProperty
    def str(self):
        return self.data.get('source', None)

    @ListProperty
    def lemma(self):
        return [s.get('lemma', None) for s in self.data.get('synthesis', [])]

    @ListProperty
    def lemma_detail(self):
        return flatten_lst([s.lemma_detail for s in self.tokens])

    @ListProperty
    def synthesis(self):
        return self.data.get('synthesis', [])

    @ListProperty
    def pos(self):
        return [s.get('tag', None) for s in self.data['synthesis']]

    @ListProperty
    def pos_detail(self):
        return flatten_lst([s.pos_detail for s in self.tokens])

    @ListProperty
    def dep(self):
        return [s.get('dep', None) for s in self.data['synthesis']]

    @DictProperty
    def language(self):
        return self.data.get('language_used', {}).get('sentence_level', {}).get('label', {})

    @ListProperty
    def meaning(self):
        return [[(m.get('super', ''), m.get('sub', '')) for m in t.get('meaning', [])] for t in self.data['synthesis']]

    @ListProperty
    def emotion(self):
        return [(e.get('type', None), round(e.get('value', 0), 4)) for e in self.data.get('ml_emotion', {}).get('sentence', [])]

    @DictProperty
    def emotion_ml(self):
        return self.emotion

    @DictProperty
    def sentiment_ml(self):
        return self.data.get('ml_sentiment', {}).get('sentence', {}).get('value', 0)

    @DictProperty
    def sentiment(self):
        return self.data.get('sentiment', {}).get('values', {})

    @ListProperty
    def sentiment_target(self):
        return [(e.get('target', {}).get('source', None), e.get('source', {}).get('source', None), e.get('value', 0)) for e in self.data.get('sentiment', {}).get('elements', []) if 'target' in e and e['target']]

    @StrProperty
    def sentence_type(self):
        return self.data.get('sentence_acts', {}).get('predict', None)

    @ListProperty
    def ner(self):
        return [t.ner for t in self.tokens]

    @ListProperty
    def morphology(self):
        return [t.morphology for t in self.tokens]

    @ListProperty
    def coreference(self):
        return [t.coreference for t in self.tokens]

class Document(TextChunk):
    __slots__ = ("sentences", "data", "n", "max", "id")
    next_id = 0
    def __init__(self, sentences, id=None):
        super(Document, self).__init__()
        self.sentences = [Sentence(s, i) for i, s in enumerate(sentences)]
        self.max = len(self.sentences)
        self.data = self.sentences
        if id is None:
            id = Document.next_id
            Document.next_id += 1
        self.id = str(id)

    @property
    def idx(self):
        return self.id

    @ListProperty
    def subsentences(self):
        return flatten_lst([s.subsentences for s in self.sentences])

    @ListProperty
    def tokens(self):
        return flatten_lst([s.tokens for s in self.sentences])

    @ListProperty
    def documents(self):
        return [self]

    def add_sentence(self, sentence):
        self.sentences.append(sentence)
        self.max += 1

    def __repr__(self):
        return str(self.sentences)

    def _get_data(self):
        return [s.data for s in self.sentences]

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < self.max:
            self.n += 1
            return self.sentences[self.n - 1]
        else:
            raise StopIteration

class NLP(TextChunk):
    """ Class for data analysis of API return.
        Takes a Client class as input which is used to make api requests.
        Provides both high and low level methods to access data via specific class
        for each key in the api result or methods designed for specific use cases.
        """
    def __init__(self, api_key = None, client = None, data = None, no_print=False):
        super(NLP, self).__init__()
        self.client = None
        if client or api_key:
            self.add_client(client, api_key)
        self.documents = []
        self.max = len(self.documents)
        self.data = self.documents
        self.fields = [p for p in dir(Sentence) if isinstance(getattr(Sentence,p),property)]
        if 'token_flat' not in self.fields:
            self._generate_properties()

        def doNothing(*args):
            pass
        global print
        if no_print:
            print = doNothing

    @ListProperty
    def sentences(self):
        """ Direct access to all sentences"""
        return flatten_lst([d.sentences for d in self.documents])

    @ListProperty
    def subsentences(self):
        """ Direct access to all subsentences"""
        sentences = flatten_lst([d.sentences for d in self.documents])
        return flatten_lst([s.subsentences for s in sentences])

    @ListProperty
    def tokens(self):
        """ Direct access to all tokens"""
        return flatten_lst([s.tokens for s in self.sentences])

    @property
    def nlp(self):
        return [self]

    def _generate_properties(self):
        """ Takes properties of the Sentence class and dynamically create properties for 
            NLP, Document and Subsentence class."""
        for _class in [Document, NLP]: #adding to nlp and document
            for field in self.fields:
                if field in ['subsentences', 'tokens', 'sentences']:
                    continue
                setattr(_class, field, property(self._make_lambda(field)))
                setattr(_class, field + '_flat', property(self._make_lambda(field, True)))
        for _class in [Subsentence]: #adding specific properties for subsentence
            for field in [p for p in dir(Sentence) if isinstance(getattr(Sentence, p), property)]:
                filters = ['sentiment_target', 'morphology', 'ner', 'token', 'str', \
                            'lemma', 'lemma_detail', 'synthesis', 'pos', 'pos_detail', 'dep', 'language', 'meaning', \
                            'sentiment', 'sentiment_ml', 'emotion', 'emotion_ml', 'tokens']
                if field in filters:
                    setattr(_class, field, getattr(Sentence, field))
                elif field not in ['subsentences']:
                    setattr(_class, field, property(self._make_lambda_sub(field)))
        for _class in [Sentence, Subsentence]: #adding flat variants to sentence and subsentence for convenience
            for field in self.fields:
                setattr(_class, field + '_flat', property(self._make_lambda_sent(field)))

    def _make_lambda(self, field, flatten = False):
        """ Returns lambda functions to be added as property to classes"""
        if not flatten:
            return lambda c_self : [getattr(s, field) for s in c_self.data]
        else:
            return lambda c_self : flatten_lst([getattr(s, field) for s in c_self.data])

    def _make_lambda_sub(self, field):
        """ Returns lambda functions to be added as property to classes"""
        return lambda x: field + ' not available for subsentences'

    def _make_lambda_sent(self, field):
        """ Returns lambda functions to be added as property to classes"""
        return lambda c_self: flatten_lst(getattr(c_self, field))

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < self.max:
            self.n += 1
            return self.documents[self.n - 1]
        else:
            raise StopIteration

    def _check_result(self, result):
        """ Check if synthesis key is empty """
        bad_idx = []
        for i, r in enumerate(result):
            if not 'synthesis' in r:
                bad_idx.append(i)
        return bad_idx

    def _request(self, data, skip_document = False):
        ''' Input: string or list of string'''
        results = []
        
        for i, seq in enumerate(data):
            try:
                res = self.client.request(seq)
                if res is None:
                    raise RequestError
                results += res
            except RequestError:
                if skip_document:
                    print("WARNING: Skipping document, request error ")
                    return None
                else:
                    print("WARNING request error: Document " + str(Document.next_id) + " skipping sentence " + str(i))

        try:
            bad_idx = self._check_result(results)
            if bad_idx:
                if skip_document:
                    raise PipelineError
                else:
                    for idx in sorted(bad_idx, reverse=True):
                        print("WARNING pipeline error: Document " + str(Document.next_id) + " skipping sentence " + str(idx))
                        results.pop(idx)
        except PipelineError:
            print("WARNING: Skipping document, pipeline error ")
            return None
        return results

    def add_client(self, client = None, api_key = None):
        if client and isinstance(client, Client):
            self.client = client
        elif api_key:
            self.client = Client(api_key)
        else:
            print('Please provide a client instance or an api_key.')

    def add_document(self, document, skip_document=False, id=None):
        """ Performs request to lettria API for a document and stores it.
            skip_document: If False document is added even if empty or request failed.
            id: Id given to document, by default sequential integer """
        
        def preprocess_input(document):
            if isinstance(document, str):
                if document.strip():
                    document = [document.strip()]
                else:
                    return None
            elif isinstance(document, list):
                document = [d.strip() for d in flatten_lst(document) if d.strip()]
            else:
                print('document argument should be of type list or str.')
                raise TypeError
            return document
        
        if not self.client:
            print('Failure : Please assign a client or api_key with add_client() in order to make a request. ')
            return
        
        results = None
        document = preprocess_input(document)
        
        if document:
            results = self._request(document, skip_document=skip_document)

        if results is None and skip_document == True:
            if not document:
                print("Skpping document, received empty input.")
            else:
                print("Skpping document, processing failed.")
        else:
            if isinstance(results, list):
                self.documents.append(Document(results, id=id))
                print("Added document " + str(self.documents[-1].id) + '.')
            else:
                if not document:
                    self.documents.append(Document([], id=id))
                    print("Added empty document " + str(self.documents[-1].id) + ': received empty input.')
                else:
                    self.documents.append(Document([], id=id))
                    print("Added empty document " + str(self.documents[-1].id) + ': processing failed.')
            self.max += 1

    def add_document_data(self, doc_data, id=None):
        self.documents.append(Document(doc_data, id))
        self.max += 1

    def _get_data(self):
        return [d._get_data() for d in self.documents]

    def save_result(self, *args):
        """ Alias for save_result"""
        self.save_results(*args)

    def save_results(self, file = ''):
        """ Writes json result to a file with the specified name."""
        path_ok = 0
        c = 0
        if not self.data:
            print("No data to save.")
            return
        if not file:
            file = 'results'
            while not path_ok:
                path = file + '_' + str(c) + '.jsonl'
                if not os.path.isfile(path):
                    path_ok = 1
                else:
                    c += 1
        else:
            path = file
            if path.endswith('.jsonl'):
                path = path[:-6]
            path = path + '.jsonl'
        try:
            with jsonl.open(path, 'w') as fw:
                for d in self.documents:
                    fw.write({'document_id':d.id, 'data':d._get_data()})
                # json.dump({'document_ids': [d.id for d in self.documents],'documents':self._get_data()}, f)
            print(f'Results saved to {path}')
        except Exception as e:
            print(e)

    def load_result(self, *args):
        """ Alias for load_results"""
        self.load_results(*args)

    def load_results(self, path = 'results_0', reset = False, chunksize = None):
        """ Loads result from a valid json file."""
        if path.endswith('.json') or path.endswith('.jsonl'):
            pass
        else:
            path = path + '.jsonl'
        try:
            if reset:
                self.reset_data()
            if path.endswith('jsonl'):
                with jsonl.open(path, 'r') as f:
                    for line in f:
                        self.add_document_data(line.get('data'), id=line.get('document_id', None))
            elif path.endswith('json'):
                with open(path, 'r') as f:
                    result = json.load(f)
                    if isinstance(result, dict):
                        assert len(result.get('document_ids', [])) == len(result.get('documents', [])), \
                                "'document_ids' and 'documents' should be of similar length"
                        for id_, r in zip(result['document_ids'], result['documents']):
                            self.add_document_data(r, id=id_)
                    else:
                        for r in result:
                            self.add_document_data(r)
            print(f'Loaded {path} successfully')
        except Exception as e:
            print('Failure to load ' + str(path) + ': ')
            print(e, '\n')

    def split_results(self, input_file, max_document_per_file, output_file=''):
        """ Splits existing json results into multiple files """
        if not output_file:
            output_file = input_file
        with open(input_file, 'r') as f:
            results = json.load(f)
        i = 0
        count = 0
        while i < len(self.documents):
            if isinstance(results, list):
                with open(f"{count}_{output_file}", 'w') as fw:
                    json.dump(result[i:i + max_document_per_file], fw)
            elif isinstance(results, dict):
                with open(f"{count}_{output_file}", 'w') as fw:
                    json.dump({'document_ids':results['document_ids'][i:i + max_document_per_file],\
                             'documents':results['documents'][i:i + max_document_per_file]}, fw)
            i += max_document_per_file
            count += 1
            results['document_ids']
        # for n in number:

    def reset_data(self):
        """ Erase current data """
        self.documents = []
        self.data = self.documents
        self.max = 0
        Document.next_id = 0


from .NLP import NLP
from .utils import flatten_lst

GLOBAL = ['g', 'global', 'glob']
DOC = ['d', 'doc', 'document', 'documents']
SENT = ['s', 'sentence', 'sent', 'sentences']
SUB = ['sub', 'subsentence', 'subsentences']
TOK = ['t', 'token', 'tok', 'tokens']
POSITIVE = ['positive', 'pos', '+']
NEGATIVE = ['negative', 'neg', '-']
NEUTRAL = ['neutral', 'neut']

## DEPRECATED, Functional interface should be used instead.
class Sentiment:
    def __init__(self, nlp):
        if not isinstance(nlp, NLP):
            print('Please provide an instance of the NLP class')
            raise TypeError
        self.nlp = nlp

    def get_emotion(self, level='document'):
        """ Returns emotion results at the specified hierarchical level """
        emotions = []
        if level in GLOBAL:
            _iter = [self.nlp]
        elif level in DOC:
            _iter = self.nlp.documents
        elif level in SENT:
            _iter = self.nlp.sentences
        elif level in SUB:
            _iter = self.nlp.subsentences
        else:
            print("'" + str(level) + "' is an incorrect value for level argument.")
            return []
        
        for d in _iter:
            tmp = {}
            for e in d.emotion_flat:
                if e[0] not in tmp:
                    tmp[e[0]] = {'sum':0, 'occurences':0, 'average':0}
                tmp[e[0]]['sum'] += round(e[1], 4)
                tmp[e[0]]['occurences'] += 1
            for e in tmp:
                if tmp[e]['occurences'] != 0:
                    tmp[e]['average'] = round(tmp[e]['sum'] / tmp[e]['occurences'], 4)
            emotions.append(tmp)
        return emotions

    def get_sentiment(self, level='document'):
        """ Returns sentiment results at the specified hierarchical level """
        sentiments = []
        if level in GLOBAL:
            _iter = [self.nlp]
        elif level in DOC:
            _iter = self.nlp.documents
        elif level in SENT:
            _iter = self.nlp.sentences
        elif level in SUB:
            _iter = self.nlp.subsentences
        else:
            print("'" + str(level) + "' is an incorrect value for level argument.")
            return []
        
        for d in _iter:
            tmp = {}
            for e in d.sentiment_flat:
                for pol in e:
                    if pol not in tmp:
                        tmp[pol] = {'sum':0, 'occurences':0, 'average':0}
                    tmp[pol]['sum'] += e[pol]
                    if e[pol] != 0 or pol == 'total':
                        tmp[pol]['occurences'] += 1
            for e in tmp:
                if tmp[e]['occurences'] != 0:
                    tmp[e]['average'] = round(tmp[e]['sum'] / tmp[e]['occurences'], 4)
            for p in tmp:
                tmp[p]['sum'] = round(tmp[p]['sum'], 4)
            sentiments.append(tmp)
        return sentiments

    def add_dict(self, d1, d2):
        for k,v in d2.items():
            d1[k] = d1.get(k, 0) + v
        return d1

    def sum_dicts(self, lst):
        merge = {}
        for d in lst:
            merge = self.add_dict(merge, d)
        return merge

    def average_dicts(self, lst):
        merge = {}
        for d in lst:
            merge = self.add_dict(merge, d)
        c = len(lst)
        return {k:round(v/c, 5) for k,v in merge.items()}

    def word_sentiment(self, granularity = 'sentence', lemma = False, filter_pos = None):
        """ Returns sentiment associated with word"""
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        tokens = self.nlp.vocabulary(filter_pos=filter_pos, lemma=lemma)
        res = {t:{'val':0, 'occ':0} for t in tokens}
        iter_ = self.nlp.subsentences if granularity in SUB else self.nlp.sentences
        for sentence in iter_:
            val = sentence.sentiment.get('total', 0)
            tmp_iter = sentence.lemma if lemma else sentence.token
            for t, p in zip(tmp_iter, sentence.pos):
                if not filter_pos or p in filter_pos:
                    res[(t,p)]['val'] += val
                    res[(t,p)]['occ'] += 1
        res = {t: round(res[t]['val'] / (res[t]['occ'] + 1e-8), 4) for t in res.keys()}
        return res

    def meaning_sentiment(self, granularity='sentence', filter_meaning=None):
        """ Returns sentiment associated with meaning"""
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        if isinstance(filter_meaning, str):
            filter_meaning = [filter_meaning]
        meanings = list(set([m[1] for m in self.nlp.meaning_flat if m[1]]))
        if filter_meaning and isinstance(filter_meaning, list):
            meanings = filter_meaning
        res = {t:{'val':0, 'occ':0} for t in meanings}
        iter_ = self.nlp.subsentences if granularity == 'subsentence' else self.nlp.sentences
        for sentence in iter_:
            val = sentence.sentiment.get('total',0)
            tmp_iter = list(set([m[1] for m in sentence.meaning_flat if m[1]]))
            for m in tmp_iter:
                if not filter_meaning or m in filter_meaning:
                    res[m]['val'] += val
                    res[m]['occ'] += 1
        res = {t: round(res[t]['val'] / (res[t]['occ'] + 1e-8), 4) for t in res.keys()}
        return res

    def filter_polarity(self, polarity, granularity='sentence'):
        if polarity not in POSITIVE + NEGATIVE + NEUTRAL:
            print("polarity argument should be 'neutral', 'positive' or 'negative'.")
            return None
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        def check_polarity(value, polarity):
            if polarity in NEUTRAL:
                return value == 0
            elif polarity in POSITIVE:
                return value > 0
            elif polarity in NEGATIVE:
                return value < 0
            else:
                return False
        iter_ = self.nlp.subsentences if granularity == 'subsentence' else self.nlp.sentences
        return [s for s in iter_ if check_polarity(s.sentiment.get('total', 0), polarity)]

    def filter_emotion(self, emotions, granularity='sentence'):
        if isinstance(emotions, str):
            emotions = [emotions]
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        def check_emotion(emotion_lst, filter):
            emotions = [e[0] for e in emotion_lst]
            for e in emotions:
                if e in filter:
                    return True
            return False
        iter_ = self.nlp.subsentences if granularity in SUB else self.nlp.sentences
        return [s for s in iter_ if check_emotion(s.emotion, emotions)]



from .transform_morph import transform_data
from .utils import StrProperty, ListProperty, DictProperty
from .TextChunk import TextChunk

class Subsentence(TextChunk):
    __slots__ = ("data", "n", "max")

    def __init__(self, data_sentence):
        super(Subsentence, self).__init__()
        self.data = data_sentence
        self.max = len(self.data['synthesis'])

    def __repr__(self):
        return self.str

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < self.max:
            self.n += 1
            return Token(self.data['synthesis'][self.n - 1])
        else:
            raise StopIteration
    
    @ListProperty
    def subsentences(self):
        return [self]

class Token:
    __slots__ = ("data", "idx")

    def __init__(self, data, idx):
        self.data = data
        self.idx = idx

    def __repr__(self):
        return self.str

    def __str__(self):
        return self.data.get('source', None)

    @StrProperty
    def token(self):
        return self.data.get('source', None)

    @StrProperty
    def source(self):
        return self.data.get('source', None)

    @StrProperty
    def str(self):
        return self.data.get('source', None)

    @StrProperty
    def pos(self):
        return self.data.get('tag', None)

    @StrProperty
    def pos(self):
        return self.data.get('tag', None)

    @ListProperty
    def pos_detail(self):
        if self.data['nlp']:
            return [l.get('tag', None) for l in self.data['nlp']]
        else:
            return [self.data.get('tag', None)]

    @DictProperty
    def ner(self):
        type_ = self.data.get('type', '')
        value = self.data.get('value', '')
        tmp = {}
        if type_:
            tmp['type'] = type_
        if value:
            tmp['value'] = value
        return tmp

    @StrProperty
    def coreference(self):
        if 'coreference' in self.data and self.data['coreference']:
            return self.data['coreference'][0].get('source', None)
        return None

    @StrProperty
    def lemma(self):
        return self.data.get('lemma', None)

    @ListProperty
    def lemma_detail(self):
        lemmas = []
        if self.data.get('nlp', []):
            for l in self.data.get('nlp', []):
                if 'number' in l.get('lemmatizer', {}):
                    lemmas.append(l.get('lemmatizer', {}).get('number', {}))
                elif 'source' in l.get('lemmatizer', {}):
                    lemmas.append(l.get('lemmatizer', {}).get('source', {}))
                else:
                    lemmas.append(l.get('source', ''))
            return lemmas
        else:
            return [self.data.get('lemma', None)]

    @ListProperty
    def morphology(self):
        if self.data.get('nlp', []):
            self.data['nlp']
            return transform_data(self.data['nlp'])
        else:
            return []

    @StrProperty
    def dep(self):
        return self.data.get('dep', None)

    @StrProperty
    def ref(self):
        return self.data.get('ref', None)

    @ListProperty
    def meaning(self):
        return [(m.get('super', ''), m.get('sub', '')) for m in self.data.get('meaning', [])]

    def __format__(self, format_spec):
        return format(str(self), format_spec)



from .NLP import NLP
import jsonlines as jsonl

class JSONLReader:
    def __init__(self, path, chunksize = 1):
        self.path = path
        self.chunksize = chunksize
        try:
            self.io = jsonl.open(path, 'r')
            self.io_iter = self.io.iter()
        except Exception as e:
            print("Failed to open file " + str(path) + ": " + str(e))
            self.io = None
        self.io_over = False

    def __iter__(self):
        self.nrows = 0
        return self

    def __next__(self):
        if self.io_over:
            raise StopIteration
        i = 0
        nlp = NLP()
        while i < self.chunksize:
            try:
                line = next(self.io_iter)
                nlp.add_document_data(line.get('data'), id=line.get('document_id', None))
            except StopIteration:
                self.io_over = True
                self.io.close()
                if i == 0:
                    raise StopIteration
                else:
                    return nlp
            i+=1
        return nlp

def load_result(self, *args):
    """ Alias for load_results"""
    load_results(*args)

def load_results(path = 'results_0', reset = False, chunksize = None):
    """ Loads result from a valid json file."""
    if not (path.endswith('.json') or path.endswith('.jsonl')):
        path = path + '.jsonl'
    nlp = NLP()
    if not (isinstance(chunksize, int) and chunksize >= 1):
        try:
            if reset:
                self.reset_data()
            if path.endswith('jsonl'):
                with jsonl.open(path, 'r') as f:
                    for line in f:
                        nlp.add_document_data(line.get('data'), id=line.get('document_id', None))
            elif path.endswith('json'):
                with open(path, 'r') as f:
                    result = json.load(f)
                    if isinstance(result, dict):
                        assert len(result.get('document_ids', [])) == len(result.get('documents', [])), \
                                "'document_ids' and 'documents' should be of similar length"
                        for id_, r in zip(result['document_ids'], result['documents']):
                            nlp.add_document_data(r, id=id_)
                    else:
                        for r in result:
                            nlp.add_document_data(r)
            print(f'Loaded {path} successfully')
        except Exception as e:
            print('Failure to load ' + str(path) + ': ')
            print(e, '\n')
    else:
        if not path.endswith('jsonl'):
            print("chunk loading only avaiable with jsonl files.")
            return None
        return JSONLReader(path, chunksize)
    return nlp


flatten=lambda l: sum(map(flatten,l),[]) if isinstance(l,list) else [l]

def flatten_lst(lst):
    return flatten(lst)

class StrProperty(property):
    def __get__(self, obj, objtype=None):
        try:
            if obj is None:
                return self
            if self.fget is None:
                raise AttributeError("unreadable attribute")
            return self.fget(obj)
        except:
            return None

class ListProperty(property):
    def __get__(self, obj, objtype=None):
        try:
            if obj is None:
                return self
            if self.fget is None:
                raise AttributeError("unreadable attribute")
            return self.fget(obj)
        except:
            return []

class DictProperty(property):
    def __get__(self, obj, objtype=None):
        try:
            if obj is None:
                return self
            if self.fget is None:
                raise AttributeError("unreadable attribute")
            return self.fget(obj)
        except:
            return {}


from collections import defaultdict
from .patterns.patterns_token import check_pattern

GLOBAL = ['g', 'global', 'glob']
DOC = ['d', 'doc', 'document', 'documents']
SENT = ['s', 'sentence', 'sent', 'sentences']
SUB = ['sub', 'subsentence', 'subsentences']
TOK = ['t', 'token', 'tok', 'tokens']
POSITIVE = ['positive', 'positif', 'pos', '+']
NEGATIVE = ['negative', 'negatif', 'neg', '-']
NEUTRAL = ['neutral', 'neutre', 'neut']
EMOTIONS = ['disgust', 'anger', 'fear', 'joy', 'love', 'sadness', 'surprise', 'neutral']
SENTENCE_TYPES = ['command', 'assert', 'question_open', 'question_closed']

class TextChunk:
    def __init__(self):
        self.class_name = self.__class__.__name__        

    def match_pattern(self, patterns_json, level = None, print_tree=False):
        matches = defaultdict(list)

        matches = []
        if level != None and level not in DOC + SENT + SUB:
            print("level argument is invalid.")
            return None
        if self.class_name == 'NLP':
            if level == None:
                level = 'documents'
        elif self.class_name == 'Document':
            if level == None:
                level = 'sentence'
            elif level in DOC:
                print("level argument " + level + " is not available for documents.")
                return None
        elif self.class_name == 'Sentence':
            if level == None:
                level = 'sentence'
            elif level in GLOBAL + DOC:
                print("level argument " + level + " is not available for sentences.")
                return None
        elif self.class_name == 'Subsentence':
            if level in GLOBAL + DOC + SENT:
                print("level argument " + level + " is not available for subsentences.")
                return None
            level = 'subsentence'
        try:
            if level in DOC:
                for element in self.documents:
                    tmp = {}
                    for sentence in element:
                        for pattern_name, patterns in patterns_json.items():
                            res_pattern = check_pattern(sentence, patterns, print_tree)
                            if res_pattern != []:
                                tmp[pattern_name] = tmp.get(pattern_name, []) + (res_pattern)
                    if tmp:
                        matches.append((element, tmp))
            elif level in SENT:
                for sentence in self.sentences:
                    tmp = {}
                    for pattern_name, patterns in patterns_json.items():
                        res_pattern = check_pattern(sentence, patterns, print_tree)
                        if res_pattern != []:
                            tmp[pattern_name] = tmp.get(pattern_name, []) + (res_pattern)
                    if tmp:
                        matches.append((sentence, tmp))
            elif level in SUB:
                for sentence in self.subsentences:
                    tmp = {}
                    for pattern_name, patterns in patterns_json.items():
                        res_pattern = check_pattern(sentence, patterns, print_tree)
                        if res_pattern != []:
                            tmp[pattern_name] = tmp.get(pattern_name, []) + (res_pattern)
                    if tmp:
                        matches.append((sentence, tmp))
        except Exception as e:
            print("Pattern matching failed:", e)
        return matches

    def vocabulary(self, filter_pos = None, lemma=False):
        """ Generates vocabulary list of words"""
        vocabulary = []
        tokens = self.token_flat if not lemma else self.lemma_flat
        pos = self.pos_flat
        for t, p in zip(tokens, pos):
            if (t,p) not in vocabulary:
                if not filter_pos or p in filter_pos:
                    vocabulary.append((t, p))
        return vocabulary
    
    def list_entities(self):
        """ Returns dictionary of ner entities"""
        entities = {}
        for t, e in zip(self.token_flat, self.ner_flat):
            if e.get('type', None):
                for _type in e['type']:
                    if _type in entities:
                        entities[_type].append(t)
                    else:
                        entities[_type] = [t]
        return entities

    def word_count(self, filter_pos = None, lemma=False):
        """ Generates word count, document or global level """
        word_count = []
        tokens = self.token_flat if not lemma else self.lemma_flat
        pos = self.pos_flat
        word_count = {}
        for t, p in zip(tokens, pos):
            if not filter_pos or p in filter_pos:
                word_count[(t, p)] = word_count.get((t,p), 0) + 1
        return word_count

    def word_frequency(self, filter_pos = None, lemma=False):
        """ Generates frequency list of words """
        vocab = self.word_count(filter_pos=filter_pos, lemma=lemma)
        total = sum(self.word_count().values())
        return {k:round(v/(total + 1e-10), 10) for k,v in vocab.items()}

    def statistics(self):
        return {'documents': len(self.documents) if self.class_name in ['NLP'] else int(self.class_name == 'Document'),
        'sentences': len(self.sentences) if self.class_name in ['NLP', 'Document'] else int(self.class_name == 'Sentence'),
        'subsentences': len(self.subsentences) if self.class_name in ['NLP', 'Document', 'Sentence'] else int(self.class_name == 'Subsentence'),
        'tokens': len(self.tokens)}
    
    def get_emotion(self, granularity = 'sentence'):
        """ Returns emotion results at the specified hierarchical level """
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        if self.class_name == 'Subsentence':
            granularity = 'subsentence'
        emotions = {}
        _iter = self.subsentences if granularity in SUB else self.sentences
        denumerator = self.statistics().get('subsentences', 1) if granularity in SUB else self.statistics().get('sentences', 1)
        for sequence in _iter:
            for e in sequence.emotion_flat:
                if e[0] not in emotions:
                    emotions[e[0]] = {'sum':0, 'occurrences':0, 'average':0}
                emotions[e[0]]['sum'] += round(e[1], 4)
                emotions[e[0]]['occurrences'] += 1
            for e in emotions:
                if emotions[e]['occurrences'] != 0:
                    emotions[e]['average'] = round(emotions[e]['sum'] / (denumerator  + 1e-6), 4)
        for k in emotions.keys():
            emotions[k]['sum'] = round(emotions[k]['sum'], 4)
            emotions[k]['average'] = round(emotions[k]['average'], 4)
        return emotions

    def get_sentiment(self, granularity='sentence'):
        """ Returns sentiment results at the specified hierarchical level """
        class_name = self.__class__.__name__
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        if class_name == 'Subsentence':
            granularity = 'subsentence'
        _iter = self.subsentences if granularity in SUB else self.sentences
        denumerator = self.statistics().get('subsentences', 1) if granularity in SUB else self.statistics().get('sentences', 1)
        sentiments = {}
        for sequence in _iter:
            for e in sequence.sentiment_flat:
                for pol in e:
                    if pol not in sentiments:
                        sentiments[pol] = {'sum':0, 'occurrences':0, 'average':0}
                    sentiments[pol]['sum'] += e[pol]
                    if e[pol] != 0 or pol == 'total':
                        sentiments[pol]['occurrences'] += 1
            for e in sentiments:
                if sentiments[e]['occurrences'] != 0:
                    sentiments[e]['average'] = sentiments[e]['sum'] / (denumerator + 1e-6)
            for p in sentiments:
                sentiments[p]['sum'] = sentiments[p]['sum']
        for k in sentiments.keys():
            sentiments[k]['sum'] = round(sentiments[k]['sum'], 4)
            sentiments[k]['average'] = round(sentiments[k]['average'], 4)
        return sentiments
    
    def filter_polarity(self, polarity, granularity='sentence'):
        """ Returns sentence or subsentence objects with the specified sentiment polarity"""
        if isinstance(polarity, str):
            polarity = [polarity]
        for p in polarity:
            if p not in POSITIVE + NEGATIVE + NEUTRAL:
                print("polarity argument should be 'neutral', 'positive' or 'negative'.")
                return []
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return []
        class_name = self.__class__.__name__
        if class_name == 'Subsentence':
            granularity = 'subsentence'
        def check_polarity(value, polarity):
            if len(set(NEUTRAL).intersection(set(polarity))) >= 1:
                if value == 0:
                    return True
            if len(set(POSITIVE).intersection(set(polarity))) >= 1:
                if value > 0:
                    return True
            if len(set(NEGATIVE).intersection(set(polarity))) >= 1:
                if value < 0:
                    return True
            else:
                return False
        _iter = self.subsentences if granularity in SUB else self.sentences
        return [s for s in _iter if check_polarity(s.sentiment.get('total', 0), polarity)]

    def filter_emotion(self, emotions, granularity='sentence'):
        """ Returns sentence or subsentence objects with the specified emotion """
        if isinstance(emotions, str):
            emotions = [emotions]
        for p in emotions:
            if p not in EMOTIONS:
                print("Available emotions are :", ' '.join(EMOTIONS))
                return []
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        class_name = self.__class__.__name__
        if class_name == 'Subsentence':
            granularity = 'subsentence'
        def check_emotion(emotion_lst, filter):
            emotions = [e[0] for e in emotion_lst]
            for e in emotions:
                if e in filter:
                    return True
            return False
        _iter = self.subsentences if granularity in SUB else self.sentences
        return [s for s in _iter if check_emotion(s.emotion, emotions)]

    def filter_type(self, sentence_type):
        """ Returns sentence objects with the specified type"""
        if isinstance(sentence_type, str):
            sentence_type = [s_type]
        class_name = self.__class__.__name__
        if class_name == 'Subsentence':
            print("Sentence type is not available for subsentences.")
            return []
        for p in sentence_type:
            if p not in SENTENCE_TYPES:
                print("Error, available types are :", ' '.join(SENTENCE_TYPES))
                return []
        return [s for s in self.sentences if s.sentence_type in sentence_type]

    def word_sentiment(self, granularity = 'sentence', lemma = False, filter_pos = None, average=True):
        """ Returns sentiment associated with word"""
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        class_name = self.__class__.__name__
        if class_name == 'Subsentence':
            granularity = 'subsentence'
        tokens = self.vocabulary(filter_pos=filter_pos, lemma=lemma)
        _iter = self.subsentences if granularity in SUB else self.sentences
        if average:
            res = {t:{'values':0, 'occurrences':0} for t in tokens}
            for sentence in _iter:
                val = sentence.sentiment.get('total', 0)
                tmp_iter = sentence.lemma if lemma else sentence.token
                for t, p in zip(tmp_iter, sentence.pos):
                    if not filter_pos or p in filter_pos:
                        res[(t,p)]['values'] += val
                        res[(t,p)]['occurrences'] += 1
            res = {t: round(res[t]['values'] / (res[t]['occurrences'] + 1e-8), 4) for t in res.keys()}
        else:
            res = {t:[] for t in tokens}
            for sentence in _iter:
                val = sentence.sentiment.get('total', 0)
                tmp_iter = sentence.lemma if lemma else sentence.token
                for t, p in zip(tmp_iter, sentence.pos):
                    if not filter_pos or p in filter_pos:
                        res[(t,p)].append(val)
        return res

    def word_emotion(self, granularity = 'sentence', lemma = False, filter_pos = None, average=True):
        """ Returns emotion associated with word"""
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        class_name = self.__class__.__name__
        if class_name == 'Subsentence':
            granularity = 'subsentence'
        tokens = self.vocabulary(filter_pos=filter_pos, lemma=lemma)
        _iter = self.subsentences if granularity in SUB else self.sentences
        if average:
            res = {t:{} for t in tokens}
            for sentence in _iter:
                val = sentence.emotion
                tmp_iter = sentence.lemma if lemma else sentence.token
                for t, p in zip(tmp_iter, sentence.pos):
                    if not filter_pos or p in filter_pos:
                        for v in val:
                            if not v[0] in res[(t,p)]:
                                res[(t,p)][v[0]] = {'values':0, 'occurrences':0}
                            res[(t,p)][v[0]]['values'] += v[1]
                            res[(t,p)][v[0]]['occurrences'] += 1
            res = {t: {v:round(res[t][v]['values'] / (res[t][v]['occurrences'] + 1e-8), 4) for v in val} for t, val in res.items()}
        else:
            res = {t:{} for t in tokens}
            for sentence in _iter:
                val = sentence.emotion
                tmp_iter = sentence.lemma if lemma else sentence.token
                for t, p in zip(tmp_iter, sentence.pos):
                    if not filter_pos or p in filter_pos:
                        for v in val:
                            if not v[0] in res[(t,p)]:
                                res[(t,p)][v[0]] = []
                            res[(t,p)][v[0]].append(v[1])
        return res

    def meaning_sentiment(self, granularity='sentence', filter_meaning=None, average=True):
        """ Returns sentiment associated with meaning"""
        if isinstance(filter_meaning, str):
            filter_meaning = [filter_meaning]
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        class_name = self.__class__.__name__
        if class_name == 'Subsentence':
            granularity = 'subsentence'
        if filter_meaning and isinstance(filter_meaning, list):
            meanings = filter_meaning
        else:
            meanings = list(set([m[1] for m in self.meaning_flat if m[1]]))
        _iter = self.subsentences if granularity == 'subsentence' else self.sentences
        if average:
            res = {t:{'values':0, 'occurrences':0} for t in meanings}
            for sentence in _iter:
                val = sentence.sentiment.get('total',0)
                tmp_iter = list(set([m[1] for m in sentence.meaning_flat if m[1]]))
                for m in tmp_iter:
                    if not filter_meaning or m in filter_meaning:
                        res[m]['values'] += val
                        res[m]['occurrences'] += 1
            res = {t: round(res[t]['values'] / (res[t]['occurrences'] + 1e-8), 4) for t in res.keys()}
        else:
            res = {t:[] for t in meanings}
            for sentence in _iter:
                val = sentence.sentiment.get('total',0)
                tmp_iter = list(set([m[1] for m in sentence.meaning_flat if m[1]]))
                for m in tmp_iter:
                    if not filter_meaning or m in filter_meaning:
                        res[m].append(val)
        return res

    def meaning_emotion(self, granularity='sentence', filter_meaning=None, average=True):
        """ Returns emotion associated with meaning"""
        if isinstance(filter_meaning, str):
            filter_meaning = [filter_meaning]
        if granularity not in SENT + SUB:
            print("granularity argument should be 'sentence' or 'subsentence'")
            return None
        class_name = self.__class__.__name__
        if class_name == 'Subsentence':
            granularity = 'subsentence'
        if filter_meaning and isinstance(filter_meaning, list):
            meanings = filter_meaning
        else:
            meanings = list(set([m[1] for m in self.meaning_flat if m[1]]))
        _iter = self.subsen2tences if granularity == 'subsentence' else self.sentences
        if average:
            res = {t:{} for t in meanings}
            for sentence in _iter:
                val = sentence.emotion
                tmp_iter = list(set([m[1] for m in sentence.meaning_flat if m[1]]))
                for m in tmp_iter:
                    if not filter_meaning or m in filter_meaning:
                        for v in val:
                            if not v[0] in res[m]:
                                res[m][v[0]] = {'values':0, 'occurrences':0}
                            res[m][v[0]]['values'] += v[1]
                            res[m][v[0]]['occurrences'] += 1
            res = {t: {v:round(res[t][v]['values'] / (res[t][v]['occurrences'] + 1e-8), 4) for v in val} for t, val in res.items()}
        else:
            res = {t:{} for t in meanings}
            for sentence in _iter:
                val = sentence.emotion
                tmp_iter = list(set([m[1] for m in sentence.meaning_flat if m[1]]))
                for m in tmp_iter:
                    if not filter_meaning or m in filter_meaning:
                        for v in val:
                            if not v[0] in res[m]:
                                res[m][v[0]] = []
                            res[m][v[0]].append(v[1])
        return res



import requests

class Client:
	def __init__(self, key=None):
		if not key:
			raise Exception('Missing key for initialization')
		if isinstance(key, str) is False:
			raise Exception('Wrong key format')
		if key.startswith('LettriaProKey'):
			key = key[14:]
		self.key = key
		self.headers = { 'Authorization': 'LettriaProKey ' + str(self.key), 'content-type': 'application/json' }
		self.max_try = 3

	def print_response_error(self, response):
		if 'Error' in response:
			print("Lettria SDK: \033[1;31;40mERROR:\033[0;37;40m {}".format(response['Error']))
		else:
			print(response)

	def server_request(self, text):
		result = None
		response = None
		i = 0
		while i < self.max_try:
			try:
				response = requests.post('https://api.lettria.com/main', headers=self.headers, json={'text' : text}).json()
				if response and not isinstance(response, list):
					raise Exception
				result = response
				break
			except Exception as e:
				i += 1
		if result is None:
			print(f'Request failed after {self.max_try} tries.')
		if result and not isinstance(result, list):
			result = None
		return result

	def request(self, text):
		sentences_json = self.server_request(text)
		return sentences_json



# coding: utf-8

from setuptools import setup, find_packages

setup(
    name="lettria",
    version="5.2.1",
    packages=find_packages(),
    description="Lettria official SDK for python",
    long_description="Lettria official SDK for python",
    author="Maxence",
    author_email="maxence@lettria.com",
    url="https://github.com/Lettria/sdk-python",
    license="MIT",
    install_requires=[
      "requests>=2.10",
      "jsonlines"
    ],
    keywords="lettria french nlp taln ner nlu chatbot bigdata text analysis sentiment emotion"
)



from .attribute_checker import compare_attr
import copy
from .DepTree import DepTree

dep_ops = {
    '>' : lambda x: x.children,
    '>>' : lambda x: x.descendants,
    '<' : lambda x: [x.parent] if x.parent else [],
    '<<' : lambda x: x.parents,
    '.' : lambda x: [x.next],
    '.*' : lambda x: x.nexts,
    ';' : lambda x: [x.prev],
    ';*' : lambda x: x.prevs,
    '$+' : lambda x: [k for k in x.siblings if x.idx == k.idx - 1],
    '$-' : lambda x: [k for k in x.siblings if x.idx == k.idx + 1],
    '$++' : lambda x: x.right_siblings,
    '$--' : lambda x: x.left_siblings,
}

# A < B        A is the immediate dependent of B.
# A > B        A is the immediate head of B.
# A << B    A is the dependent in a chain to B following dep → head paths.
# A >> B    A is the head in a chain to B following head → dep paths.
# A . B        A immediately precedes B, i.e. A.i == B.i - 1, and both are within the same dependency tree.
# A .* B    A precedes B, i.e. A.i < B.i, and both are within the same dependency tree (not in Semgrex).
# A ; B        A immediately follows B, i.e. A.i == B.i + 1, and both are within the same dependency tree (not in Semgrex).
# A ;* B    A follows B, i.e. A.i > B.i, and both are within the same dependency tree (not in Semgrex).
# A $+ B    B is a right immediate sibling of A, i.e. A and B have the same parent and A.i == B.i - 1.
# A $- B    B is a left immediate sibling of A, i.e. A and B have the same parent and A.i == B.i + 1.
# A $++ B    B is a right sibling of A, i.e. A and B have the same parent and A.i < B.i.
# A $-- B    B is a left sibling of A, i.e. A and B have the same parent and A.i > B.i.


def find_matching_nodes(step_pattern, target_node):
    op = step_pattern["REL_OP"]
    result = []
    candidates = dep_ops[op](target_node) #assume op is ok
    for c in candidates:
        if compare_attr(c.root.tokens[c.idx], step_pattern['RIGHT_ATTRS']):
            result.append(c)
    return result

def get_match_from_root(root_node, root_pattern, pattern, c_pattern = None, c_nodes = None):
    def find_pattern_by_id(pattern_list, nodes_list, id):
        for p, n in zip(pattern_list, nodes_list):
            if p['RIGHT_ID'] == id:
                return p,n
        return None, None
    current_pattern = [root_pattern] if c_pattern == None else c_pattern
    current_nodes = [root_node] if c_nodes == None else c_nodes
    if c_nodes:
        indent = '      '
    else:
        indent = ''
    result = []
    j = 0
    while len(current_pattern) != len(pattern):
        if j < len(pattern):
            step_pattern = pattern[j]
        else:
            break
        if step_pattern in current_pattern:
            j += 1
            continue
        target_pattern, target_node = find_pattern_by_id(current_pattern, current_nodes, step_pattern['LEFT_ID'])
        if target_pattern == None:
            j += 1
            continue
        tmp_match = find_matching_nodes(step_pattern, target_node)
        if tmp_match:
            current_pattern.append(step_pattern)
            if len(tmp_match) > 1:
                for t in tmp_match[1:]:
                    tmp_node_lst = current_nodes + [t]
                    if len(tmp_node_lst) == len(pattern):
                        result += [tmp_node_lst]
                    else:
                        result += get_match_from_root(root_node, root_pattern, pattern, copy.deepcopy(current_pattern), tmp_node_lst)
            tmp_match = tmp_match[0]
            current_nodes.append(tmp_match)
            j += 1
            if j != len(current_pattern):
                j = 0
        else:
            break
        if len(current_pattern) == len(pattern):
            return result + [current_nodes]
    return []

def check_pattern_dependency(sentence_data, pattern, print_tree):
    def find_root_pattern(pattern):
        for e in pattern:
            if 'LEFT_ID' not in e:
                return e
        return None
    tree = DepTree.grow_tree(sentence_data.data)
    tree.tokens = sentence_data.tokens
    if print_tree:
        tree.print_tree()
    patterns_caught = []
    root_pattern = find_root_pattern(pattern)
    if root_pattern is None:
        print("Pattern is invalid, no root is present.")
        return []
    if tree is None:
        return []
    root_nodes = tree.find_node_attribute(root_pattern['RIGHT_ATTRS'])
    for root_node in root_nodes:
        res = get_match_from_root(root_node, root_pattern, pattern)
        if res:
            patterns_caught += res
    results = []
    for p in patterns_caught:
        results.append([k.root.tokens[k.idx] for k in p])
    return results


from .patterns_dependency import check_pattern_dependency
from .attribute_checker import compare_attr

def check_pattern(data, pattern, print_tree):
    if 'RIGHT_ID' in pattern[0] or 'LEFT_ID' in pattern[0]:
        return check_pattern_dependency(data, pattern, print_tree)
    i = 0
    j = 0
    has_matched = False
    patterns_caught = []
    current_pattern = []
    while i < len(data.tokens):
        current_token = data.tokens[i]
        current_node = pattern[j]
        op = current_node.get('OP', '.')
        if op == '.':
            if compare_attr(current_token, current_node):
                j += 1
                current_pattern.append(current_token)
                # print("ok.  ", current_token, current_node)
            else:
                # print("fail.", current_token, current_node)
                j = 0
                current_pattern = []
        elif op == '+':
            if compare_attr(current_token, current_node):
                has_matched = True
                current_pattern.append(current_token)
                # print("ok.  ", current_token, current_node)
            else:
                if has_matched == True:
                    i -= 1
                    j += 1
                    has_matched = False
                else:
                    j = 0
                    current_pattern = []
        elif op == '*':
            if compare_attr(current_token, current_node):
                has_matched = True
                current_pattern.append(current_token)
                # print("ok.  ", current_token, current_node)
            else:
                if has_matched == True:
                    i -= 1
                    j += 1
                    has_matched = False
                else:
                    i -= 1
                    j += 1
                # print("fail.", current_token, current_node)
        elif op == '!':
            if not compare_attr(current_token, current_node):
                j += 1
                current_pattern.append(current_token)
                # print("ok!  ", current_token, current_node)
            else:
                # print("fail!", current_token, current_node)
                j = 0
                current_pattern = []
        elif op == '?':
            j += 1
            if compare_attr(current_token, current_node):
                # print("ok?  ", current_token, current_node)
                current_pattern.append(current_token)
            else:
                # print("fail?", current_token, current_node)
                i -= 1
        i += 1
        if j == len(pattern):
            patterns_caught.append(current_pattern)
            current_pattern = []
            j = 0
        if i == len(data.tokens):
            break
    return patterns_caught
